	include init.s

scrambled=0

; Unscrambled:
; 320x256: 836
; 320x120: 394
; Scrambled:
; 320x256: 558
; 320x120: 264

screenw=320 ; must be multiple of 16
screenh=256
screend=4   ; fixed
bplbytes=(screenw/8)*screenh
bplwords=bplbytes/2
xstart=$81+(320-screenw)/2
ystart=$2c+(256-screenh)/2 ; first scanline

	SECTION code, code

; wait for line \1 (e.g. #$12d), note: must not be d0
; trashes d0, assumes a6=custom
waitline macro
.\@:
    	move.l  vposr(a6), d0
    	lsr.l	#8, d0
    	and.w	#$1ff, d0
    	cmp.w   \1, d0
    	bne.s   .\@
	endm

waitblit macro
	tst	dmaconr(a6)
.\@:
	btst.b	#DMAB_BLTDONE-8, dmaconr(a6)
	bne.s	.\@
	endm

; Read horizontal sync event counter
readtodh macro
        moveq   #0, \1
        move.b  $bfda00, \1
        swap    \1
        move.b  $bfd900, \1
        lsl.w   #8, \1
        move.b  $bfd800, \1
        endm


        ifeq scrambled
; Second half of the buffer can be used for temporary storage after the first pass
; If the first half is omitted (and the buffer size therefore reduced) another buffer
; of size bplbytes*4 needs to be available (in chip ram)
tempbuffer=buffer+4*bplbytes
        endc


main:
        bsr     init

        move.w  #DMAF_SETCLR!DMAF_MASTER!DMAF_BLITHOG!DMAF_BLITTER, dmacon(a6)

        readtodh d0
.r:     readtodh d7
        cmp.l   d0, d7
        beq     .r

        move.l  #buffer, a0
        move.l  #tempbuffer, a1
        bsr     c2p

        readtodh d0
        sub.l   d7, d0

        move.l  #text+40, a0
        bsr     convertnum
        move.l  a0, a1
        move.l  #buffer, a0
        bsr     writestr

        move.w  #16-1, d0
        lea     color(a6), a0
        move.l  #palette, a1
.pal:
        move.w  (a1)+, (a0)+
        dbf     d0, .pal

        waitline #10
        move.l  #copperlist, cop1lc(a6)
        clr.w   copjmp1(a6)
        moveq   #0, d0
        move.w  d0, bpl1mod(a6)
        move.w  d0, bpl2mod(a6)
        move.w  d0, bplcon1(a6)
        move.w  d0, bplcon2(a6)
        move.w  d0, bplcon3(a6)
        move.w  d0, bplcon4(a6)

	move.w 	#(ystart<<8)!xstart, diwstrt(a6)
	move.w 	#(((ystart+screenh)<<8)&$ff00)!((xstart+screenw)&$ff), diwstop(a6)
	move.w	#xstart/2-8, ddfstrt(a6)
	move.w	#xstart/2-8+8*((screenw+15)/16-1), ddfstop(a6)
        move.w  #screend<<12!$0200, bplcon0(a6)

        move.w  #DMAF_SETCLR!DMAF_MASTER!DMAF_RASTER!DMAF_COPPER!DMAF_BLITTER, dmacon(a6)

.waitmouse:
	btst	#6, ciaa
	bne.s	.waitmouse
        rts

init:
        ; Initialize copperlist
        move.l  #buffer, d0
        move.l  #copperlist, a0

        rept screend
        move.w  d0, 6(a0)
        swap    d0
        move.w  d0, 2(a0)
        swap    d0
        addq.l  #8, a0
        add.l   #bplbytes, d0
        endr

        ifne scrambled
        move.l  #buffer, a0
        move.l  a0, a1
        move.l  #bplbytes*4/2-1, d0
.cvt:
        move.w  (a1), d1
        move.w  2(a1), d2
        lsr.w   #2, d2
        or.w    d2, d1
        move.w  d1, (a0)+
        addq.l  #4, a1
        dbra    d0, .cvt
        endc ; scrambled

        rts

doblit macro
        rept (\1)>>10
        move.w  #1, bltsize(a6)
        waitblit
        endr
        if (\1)&1023
        move.w  #((\1)&1023)*64+1, bltsize(a6)
        waitblit
        endc
        endm
;
; Performs 2x1 C2P on specially prepared data.
;
; Pixel format is: %aa00bb00cc00dd00 (where a is the most significant bit of the 4 bpl pixel)
;
; The first pass merges two adjacent pixels together. If "scrambled" is non-zero it assumes
; this step has already been performed (perhaps as part of the rendering process).
; Note: Two pixels p1 and p2 (with p1 being leftmost) are combined like this: p1 | p2 >> 2
;

; There's some redundant setup done for some of the blits, and it could likely be done smarter.
; Improving this is left as an exercise for the reader :)


; Input: a0=chunky buffer (format as above), a1=tempbuffer (can be buffer+4*bplbytes if not pre-scrambled)
; Output: buffer converted to planar mode (format as described above)
; Transhes: a2
c2p:
        ; Common to all passes
        move.l  #-1, bltafwm(a6)

        ifeq scrambled

        ;
        ; First pass (omitted if adjacent pixels have already been combined)
        ;

        move.l  a0, bltapt(a6)
        lea     2(a0), a2
        move.l  a2, bltbpt(a6)
        move.l  a0, bltdpt(a6)
        move.w  #2, bltamod(a6)
        move.w  #2, bltbmod(a6)
        move.w  #0, bltdmod(a6)
        move.w  #SRCA!SRCB!DEST!$FC, bltcon0(a6)
        move.w  #2<<12, bltcon1(a6)
        doblit  4*bplwords

        endc ; scrambled

        ;
        ; Second pass
        ;

        move.l  a0, bltapt(a6)
        lea     2(a0), a2
        move.l  a2, bltbpt(a6)
        move.l  a1, bltdpt(a6)
        move.w  #$f0f0, bltcdat(a6)
        move.w  #6, bltamod(a6)
        move.w  #6, bltbmod(a6)
        move.w  #6, bltdmod(a6)
        move.w  #SRCA!SRCB!DEST!$E4, bltcon0(a6)
        move.w  #4<<12, bltcon1(a6)
        doblit  bplwords

        lea     4(a0), a2
        move.l  a2, bltapt(a6)
        addq.l  #2, a2
        move.l  a2, bltbpt(a6)
        lea     2(a1), a2
        move.l  a2, bltdpt(a6)
        move.w  #$f0f0, bltcdat(a6)
        move.w  #6, bltamod(a6)
        move.w  #6, bltbmod(a6)
        move.w  #6, bltdmod(a6)
        move.w  #SRCA!SRCB!DEST!$E4, bltcon0(a6)
        move.w  #4<<12, bltcon1(a6)
        doblit  bplwords

        move.l  a0, a2
        add.l   #4*bplbytes-8, a2
        move.l  a2, bltapt(a6)
        addq.w  #2, a2
        move.l  a2, bltbpt(a6)
        move.l  a1, a2
        add.l   #4*bplbytes-4, a2
        move.l  a2, bltdpt(a6)
        move.w  #$f0f0, bltcdat(a6)
        move.w  #6, bltamod(a6)
        move.w  #6, bltbmod(a6)
        move.w  #6, bltdmod(a6)
        move.w  #4<<12!SRCA!SRCB!DEST!$E4, bltcon0(a6)
        move.w  #BLITREVERSE, bltcon1(a6)
        doblit  bplwords

        move.l  a0, a2
        add.l   #4*bplbytes-4, a2
        move.l  a2, bltapt(a6)
        addq.l  #2, a2
        move.l  a2, bltbpt(a6)
        move.l  a1, a2
        add.l   #4*bplbytes-2, a2
        move.l  a2, bltdpt(a6)
        move.w  #$f0f0, bltcdat(a6)
        move.w  #6, bltamod(a6)
        move.w  #6, bltbmod(a6)
        move.w  #6, bltdmod(a6)
        move.w  #4<<12!SRCA!SRCB!DEST!$E4, bltcon0(a6)
        move.w  #BLITREVERSE, bltcon1(a6)
        doblit  bplwords

        ;
        ; Third (and final) pass
        ;

        move.l  a1, bltapt(a6)
        lea     2(a1), a2
        move.l  a2, bltbpt(a6)
        move.l  a0, a2
        add.l   #3*bplbytes, a2
        move.l  a2, bltdpt(a6)
        move.w  #$ff00, bltcdat(a6)
        move.w  #6, bltamod(a6)
        move.w  #6, bltbmod(a6)
        move.w  #0, bltdmod(a6)
        move.w  #SRCA!SRCB!DEST!$E4, bltcon0(a6)
        move.w  #8<<12, bltcon1(a6)
        doblit  bplwords

        lea     4(a1), a2
        move.l  a2, bltapt(a6)
        addq.l  #2, a2
        move.l  a2, bltbpt(a6)
        move.l  a0, a2
        add.l   #2*bplbytes, a2
        move.l  a2, bltdpt(a6)
        move.w  #$ff00, bltcdat(a6)
        move.w  #6, bltamod(a6)
        move.w  #6, bltbmod(a6)
        move.w  #0, bltdmod(a6)
        move.w  #SRCA!SRCB!DEST!$E4, bltcon0(a6)
        move.w  #8<<12, bltcon1(a6)
        doblit  bplwords

        move.l  a1, a2
        add.l   #4*bplbytes-8, a2
        move.l  a2, bltapt(a6)
        addq.l  #2, a2
        move.l  a2, bltbpt(a6)
        move.l  a0, a2
        add.l   #2*bplbytes-2, a2
        move.l  a2, bltdpt(a6)
        move.w  #$ff00, bltcdat(a6)
        move.w  #6, bltamod(a6)
        move.w  #6, bltbmod(a6)
        move.w  #0, bltdmod(a6)
        move.w  #8<<12!SRCA!SRCB!DEST!$E4, bltcon0(a6)
        move.w  #BLITREVERSE, bltcon1(a6)
        doblit  bplwords

        move.l  a1, a2
        add.l   #4*bplbytes-4, a2
        move.l  a2, bltapt(a6)
        addq.l  #2, a2
        move.l  a2, bltbpt(a6)
        move.l  a0, a2
        add.l   #1*bplbytes-2, a2
        move.l  a2, bltdpt(a6)
        move.w  #$ff00, bltcdat(a6)
        move.w  #6, bltamod(a6)
        move.w  #6, bltbmod(a6)
        move.w  #0, bltdmod(a6)
        move.w  #8<<12!SRCA!SRCB!DEST!$E4, bltcon0(a6)
        move.w  #BLITREVERSE, bltcon1(a6)
        doblit  bplwords
        rts

writestr:
.charloop:
        moveq   #0, d1
        move.b  (a1)+, d1
        beq.s   .out
        sub.b   #$20, d1
        move.l  topazdata, a3
        lea     (a3,d1.w), a3
        move.l  a0, a2
        moveq   #8-1, d1
.yloop:
        move.b  (a3), d0
        move.b  d0, 0*bplbytes(a2)
        move.b  d0, 1*bplbytes(a2)
        move.b  d0, 2*bplbytes(a2)
        move.b  d0, 3*bplbytes(a2)
        add.w   #screenw/8, a2
        add.w   #256-64, a3
        dbf     d1, .yloop
        addq    #1, a0
        bra.s   .charloop
.out:
        rts

convertnum:
        moveq   #10, d1
.cvt:
        divu.w  d1, d0
        move.l  d0, d2
        swap    d2
        add.b   #'0', d2
        move.b  d2, -(a0)
        and.l   #$ffff, d0
        bne.s   .cvt
        rts

palette:
        dc.w $222
        dc.w $542
        dc.w $245
        dc.w $444
        dc.w $E21
        dc.w $942
        dc.w $564
        dc.w $469
        dc.w $955
        dc.w $D45
        dc.w $689
        dc.w $A78
        dc.w $A84
        dc.w $5AD
        dc.w $999
        dc.w $9AC

        SECTION data_c, data_c

copperlist:
        dc.w bplpt+$00, 0
        dc.w bplpt+$02, 0
        dc.w bplpt+$04, 0
        dc.w bplpt+$06, 0
        dc.w bplpt+$08, 0
        dc.w bplpt+$0a, 0
        dc.w bplpt+$0c, 0
        dc.w bplpt+$0e, 0
        dc.l -2

buffer:
        incbin "baboon.chunky_expanded"

        SECTION bss_c, bss_c

        ifne scrambled
tempbuffer ds.b 4*bplbytes
        endc

        SECTION bss, bss

text:   ds.b 40
        even
